# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.community.load."""

import pytest
from tensorflow_datasets.core import load as core_load
from tensorflow_datasets.core.community import load


def test_import():
  builder_cls = load.builder_cls_from_module(
      'tensorflow_datasets.testing.dummy_dataset.dummy_dataset'
  )
  assert builder_cls.__name__ == 'DummyDataset'

  # Module should have been cached
  builder_cls_2 = load.builder_cls_from_module(
      'tensorflow_datasets.testing.dummy_dataset.dummy_dataset'
  )
  assert builder_cls is builder_cls_2

  from tensorflow_datasets.testing.dummy_dataset import dummy_dataset  # pylint: disable=g-import-not-at-top

  assert builder_cls is dummy_dataset.DummyDataset
  assert builder_cls.__name__ not in core_load.list_builders()


def test_bad_import():
  # Invalid module
  with pytest.raises(ImportError):
    load.builder_cls_from_module('tensorflow_datasets_invalid')

  # Valid module, but bad dataset
  with pytest.raises(ValueError, match='Could not load DatasetBuilder'):
    load.builder_cls_from_module('tensorflow_datasets')



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Scalar feature, e.g. tf.int32."""

from __future__ import annotations

from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import tensor_feature
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import type_utils


class Scalar(tensor_feature.Tensor):
  """A scalar feature of a particular dtype (e.g., np.int64 or np.float).

  Equivalent to `tfds.features.Tensor(shape=(), dtype=dtype)`.
  """

  def __init__(
      self,
      dtype: type_utils.TfdsDType,
      *,
      doc: feature_lib.DocArg = None,
  ):
    super().__init__(shape=(), dtype=dtype, doc=doc)

  @classmethod
  def from_json_content(cls, value: feature_pb2.TensorFeature) -> 'Scalar':
    return cls(dtype=feature_lib.dtype_from_str(value.dtype))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for transform."""
import functools
from typing import Iterator

import pytest
from tensorflow_datasets.core.transform import transform_lib


def add_number(number: int, increment: int) -> int:
  return number + increment


def is_even(number: int) -> bool:
  return number % 2 == 0


def duplicate(
    example: transform_lib.Example,
) -> Iterator[transform_lib.Example]:
  yield example
  yield example


def test_apply_fn_simple():
  add_two = functools.partial(add_number, increment=2)
  apply_fn = transform_lib.apply_fn(
      fn=add_two, input_feature="a", output_feature="c"
  )
  example = {"a": 1, "d": "left alone"}
  expected = [{"a": 1, "c": 3, "d": "left alone"}]
  assert list(apply_fn(example)) == expected


def test_apply_fn_nested_dict():
  add_two = functools.partial(add_number, increment=2)
  apply_fn = transform_lib.apply_fn(
      fn=add_two, input_feature="a/b", output_feature="a/c"
  )
  example = {"a": {"b": 1}, "d": "something"}
  expected = [{"a": {"b": 1, "c": 3}, "d": "something"}]
  assert list(apply_fn(example)) == expected


def test_apply_fn_sequence():
  add_two = functools.partial(add_number, increment=2)
  apply_fn = transform_lib.apply_fn(
      fn=add_two, input_feature="a", output_feature="b"
  )
  example = {"a": [1, 2, 3]}
  expected = [{"a": [1, 2, 3], "b": [3, 4, 5]}]
  assert list(apply_fn(example)) == expected


def test_apply_fn_sequence_of_dicts():
  add_two = functools.partial(add_number, increment=2)
  apply_fn = transform_lib.apply_fn(
      fn=add_two, input_feature="a/b", output_feature="a/c"
  )
  example = {"a": [{"b": [1, 2]}, {"b": [2, 3]}]}
  expected = [{"a": [{"b": [1, 2], "c": [3, 4]}, {"b": [2, 3], "c": [4, 5]}]}]
  assert list(apply_fn(example)) == expected


def test_apply_fn_different_ancestor():
  add_two = functools.partial(add_number, increment=2)
  with pytest.raises(
      ValueError, match="The out-feature must have the same ancestor"
  ):
    transform_lib.apply_fn(
        fn=add_two, input_feature="a/b", output_feature="c/d"
    )


def test_apply_filter_on_feature():
  apply_filter_fn = transform_lib.apply_filter(fn=is_even, input_feature="a")
  assert list(apply_filter_fn(example={"a": 2}))
  assert not list(apply_filter_fn(example={"a": 1}))


def test_apply_filter_on_example():
  def a_is_even(example) -> bool:
    return example["a"] % 2 == 0

  apply_filter_fn = transform_lib.apply_filter(fn=a_is_even, input_feature="")
  assert list(apply_filter_fn(example={"a": 2}))
  assert not list(apply_filter_fn(example={"a": 1}))


def test_do_fn():
  apply_do_fn = transform_lib.apply_do_fn(fn=duplicate)
  example = {"a": 1}
  assert list(apply_do_fn(example)) == [example, example]


def test_remove_feature():
  remove_fn = transform_lib.remove_feature("a")
  assert list(remove_fn(example={"a": 1, "b": 2})) == [{"b": 2}]


def test_remove_feature_multiple():
  remove_fn = transform_lib.remove_feature(["a", "b"])
  assert list(remove_fn(example={"a": 1, "b": 2, "c": 3})) == [{"c": 3}]


def test_rename_feature():
  rename_fn = transform_lib.rename_feature("a", "b")
  assert list(rename_fn(example={"a": 1})) == [{"b": 1}]


def test_rename_features():
  rename_fn = transform_lib.rename_features({"a": "x", "b": "y"})
  example = {"a": 1, "b": 2, "c": 3}
  expected = {"x": 1, "y": 2, "c": 3}
  assert list(rename_fn(example=example)) == [expected]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the Split API."""

from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import proto
from tensorflow_datasets.core import splits
from tensorflow_datasets.core.utils import shard_utils
import tensorflow_datasets.public_api as tfds

RANGE_TRAIN = list(range(0, 2000))
RANGE_TEST = list(range(3000, 3200))
RANGE_VAL = list(range(6000, 6010))


def _filename_template(
    split: str, dataset_name: str = 'ds_name', data_dir: str = '/path'
):
  return naming.ShardedFileTemplate(
      dataset_name=dataset_name,
      split=split,
      data_dir=data_dir,
      filetype_suffix='tfrecord',
  )


def test_multi_split_infos():
  split = 'train'
  split_infos = [
      tfds.core.SplitInfo(
          name=split,
          shard_lengths=[10, 10],
          num_bytes=100,
          filename_template=_filename_template(split=split, data_dir='/abc'),
      ),
      tfds.core.SplitInfo(
          name=split,
          shard_lengths=[1],
          num_bytes=20,
          filename_template=_filename_template(split=split, data_dir='/xyz'),
      ),
  ]
  multi_split_info = splits.MultiSplitInfo(name=split, split_infos=split_infos)
  assert multi_split_info.num_bytes == 120
  assert multi_split_info.num_examples == 21
  assert multi_split_info.num_shards == 3
  assert multi_split_info.shard_lengths == [10, 10, 1]
  assert multi_split_info.file_instructions == [
      shard_utils.FileInstruction(
          filename='/abc/ds_name-train.tfrecord-00000-of-00002',
          skip=0,
          take=-1,
          examples_in_shard=10,
      ),
      shard_utils.FileInstruction(
          filename='/abc/ds_name-train.tfrecord-00001-of-00002',
          skip=0,
          take=-1,
          examples_in_shard=10,
      ),
      shard_utils.FileInstruction(
          filename='/xyz/ds_name-train.tfrecord-00000-of-00001',
          skip=0,
          take=-1,
          examples_in_shard=1,
      ),
  ]
  assert (
      str(multi_split_info)
      == "MultiSplitInfo(name='train', split_infos=[<SplitInfo"
      ' num_examples=20, '
      'num_shards=2>, <SplitInfo num_examples=1, num_shards=1>])'
  )


class SplitDictTest(testing.TestCase):

  @property
  def split_dict(self):
    si = [
        tfds.core.SplitInfo(
            name='train',
            shard_lengths=[10, 10],
            num_bytes=0,
            filename_template=_filename_template('train'),
        ),
        tfds.core.SplitInfo(
            name='test',
            shard_lengths=[1],
            num_bytes=0,
            filename_template=_filename_template('test'),
        ),
    ]
    sd = splits.SplitDict(si)
    return sd

  def test_get(self):
    s = self.split_dict['train']
    self.assertEqual('train', s.name)
    self.assertEqual(2, s.num_shards)

  def test_from_proto(self):
    sd = splits.SplitDict.from_proto(
        filename_template=naming.ShardedFileTemplate(
            dataset_name='ds_name', data_dir='/path', filetype_suffix='tfrecord'
        ),
        repeated_split_infos=[
            proto.SplitInfo(name='validation', shard_lengths=[5], num_bytes=0)
        ],
    )
    self.assertIn('validation', sd)
    self.assertNotIn('train', sd)
    self.assertNotIn('test', sd)

  def test_to_proto(self):
    sd = self.split_dict
    sdp = sd.to_proto()

    # Split order is preserved
    self.assertEqual('train', sdp[0].name)
    self.assertEqual([10, 10], sdp[0].shard_lengths)

    self.assertEqual('test', sdp[1].name)
    self.assertEqual([1], sdp[1].shard_lengths)

  def test_bool(self):
    sd = splits.SplitDict([])
    self.assertFalse(sd)  # Empty split is False
    si = [
        tfds.core.SplitInfo(
            name='train',
            shard_lengths=[5],
            num_bytes=0,
            filename_template=_filename_template('train'),
        )
    ]
    sd = splits.SplitDict(si)
    self.assertTrue(sd)  # Non-empty split is True

  def test_num_shards(self):
    si = tfds.core.SplitInfo(
        name='train',
        shard_lengths=[1, 2, 3],
        num_bytes=0,
        filename_template=_filename_template(split='train'),
    )
    sd = splits.SplitDict([si])
    self.assertEqual(sd['train'].num_shards, 3)

  def test_empty_split(self):
    sd = splits.SplitDict([])
    with self.assertRaisesWithPredicateMatch(KeyError, '`splits` is empty'):
      _ = sd['train']

  def test_merge_multiple(self):
    def split_info_for(name: str, shard_lengths, template) -> splits.SplitInfo:
      return splits.SplitInfo(
          name=name,
          shard_lengths=shard_lengths,
          num_bytes=0,
          filename_template=template,
      )

    template_a = naming.ShardedFileTemplate(
        dataset_name='ds_name', data_dir='/a', filetype_suffix='tfrecord'
    )
    split_info_a1 = split_info_for('train', [1, 2, 3], template_a)
    split_info_a2 = split_info_for('test', [1], template_a)
    split_info_a3 = split_info_for('banana', [1], template_a)
    split_dict_1 = splits.SplitDict(
        [split_info_a1, split_info_a2, split_info_a3]
    )

    template_b = naming.ShardedFileTemplate(
        dataset_name='ds_name', data_dir='/b', filetype_suffix='tfrecord'
    )
    split_info_b1 = split_info_for('train', [4, 5, 6], template_b)
    split_info_b2 = split_info_for('test', [3], template_b)
    split_dict_2 = splits.SplitDict([split_info_b1, split_info_b2])

    template_c = naming.ShardedFileTemplate(
        dataset_name='ds_name', data_dir='/c', filetype_suffix='tfrecord'
    )
    split_info_c1 = split_info_for('train', [8], template_c)
    split_info_c2 = split_info_for('train', [9], template_c)
    multi_split = splits.MultiSplitInfo(
        name='train', split_infos=[split_info_c1, split_info_c2]
    )
    split_dict_3 = splits.SplitDict([multi_split])

    merged = splits.SplitDict.merge_multiple(
        [split_dict_1, split_dict_2, split_dict_3]
    )
    assert len(merged.values()) == 3
    assert merged.get('train').split_infos == [
        split_info_a1,
        split_info_b1,
        split_info_c1,
        split_info_c2,
    ]
    assert merged.get('test').split_infos == [split_info_a2, split_info_b2]
    assert merged.get('banana').split_infos == [split_info_a3]


class SplitsTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(SplitsTest, cls).setUpClass()
    tmp_dir = testing.make_tmp_dir()
    cls._builder = testing.DummyDatasetSharedGenerator(data_dir=tmp_dir)
    cls._builder.download_and_prepare()

  def test_sub_split_num_examples(self):
    s = self._builder.info.splits
    self.assertEqual(s['train[75%:]'].num_examples, 5)
    self.assertEqual(s['train[:75%]'].num_examples, 15)
    self.assertEqual(
        s['train'].num_examples,
        s['train[75%:]'].num_examples + s['train[:75%]'].num_examples,
    )

    self.assertEqual(s['test[75%:]'].num_examples, 2)
    self.assertEqual(s['test[:75%]'].num_examples, 8)
    self.assertEqual(
        s['test'].num_examples,
        s['test[75%:]'].num_examples + s['test[:75%]'].num_examples,
    )

    self.assertEqual(s['all'].num_examples, s.total_num_examples)

  def test_sub_split_file_instructions(self):
    fi = self._builder.info.splits['train[75%:]'].file_instructions
    self.assertEqual(
        fi,
        [
            shard_utils.FileInstruction(
                filename=f'{self._builder.data_dir}/dummy_dataset_shared_generator-train.tfrecord-00000-of-00001',
                skip=15,
                take=-1,
                examples_in_shard=20,
            )
        ],
    )

  def test_sub_split_num_shards(self):
    self.assertEqual(self._builder.info.splits['train[75%:]'].num_shards, 1)

  def test_sub_split_shard_lengths(self):
    self.assertEqual(
        self._builder.info.splits['train[75%:]'].shard_lengths,
        [5],
    )

  def test_sub_split_to_proto(self):
    sp = self._builder.info.splits['train[75%:]'].to_proto()
    self.assertEqual('train[75%:]', sp.name)
    self.assertEqual([5], sp.shard_lengths)

  def test_split_file_instructions(self):
    fi = self._builder.info.splits['train'].file_instructions
    self.assertEqual(
        fi,
        [
            shard_utils.FileInstruction(
                filename=f'{self._builder.data_dir}/dummy_dataset_shared_generator-train.tfrecord-00000-of-00001',
                skip=0,
                take=-1,
                examples_in_shard=20,
            )
        ],
    )

  def test_sub_split_filenames(self):
    self.assertEqual(
        self._builder.info.splits['train'].filenames,
        [
            'dummy_dataset_shared_generator-train.tfrecord-00000-of-00001',
        ],
    )
    self.assertEqual(
        self._builder.info.splits['train[75%:]'].filenames,
        [
            'dummy_dataset_shared_generator-train.tfrecord-00000-of-00001',
        ],
    )

  def test_sub_split_wrong_key(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, "Unknown split 'unknown'"
    ):
      _ = self._builder.info.splits['unknown']

  def test_split_enum(self):
    self.assertEqual(repr(splits.Split.TRAIN), "Split('train')")
    self.assertIsInstance(splits.Split.TRAIN, splits.Split)


class ReadInstructionTest(testing.TestCase):

  def setUp(self):
    super(ReadInstructionTest, self).setUp()

    self.splits = {
        'train': splits.SplitInfo(
            name='train',
            shard_lengths=[200],
            num_bytes=0,
            filename_template=_filename_template('train'),
        ),
        'test': splits.SplitInfo(
            name='test',
            shard_lengths=[101],
            num_bytes=0,
            filename_template=_filename_template('test'),
        ),
        'validation': splits.SplitInfo(
            name='validation',
            shard_lengths=[30],
            num_bytes=0,
            filename_template=_filename_template('validation'),
        ),
        'dev-train': splits.SplitInfo(
            name='dev-train',
            shard_lengths=[5, 5],
            num_bytes=0,
            filename_template=_filename_template('dev-train'),
        ),
    }

  def check_from_ri(self, ri, expected):
    res = ri.to_absolute(self.splits)
    expected_result = []
    for split_name, from_, to_ in expected:
      expected_result.append(
          splits._AbsoluteInstruction(split_name, from_, to_)
      )
    self.assertEqual(res, expected_result)
    return ri

  def check_from_spec(self, spec, expected):
    ri = splits.ReadInstruction.from_spec(spec)
    return self.check_from_ri(ri, expected)

  def assertRaises(self, spec, msg, exc_cls=ValueError):
    with self.assertRaisesWithPredicateMatch(exc_cls, msg):
      ri = splits.ReadInstruction.from_spec(spec)
      ri.to_absolute(self.splits)

  def test_valid(self):
    # Simple split:
    ri = self.check_from_spec('train', [('train', None, None)])
    self.assertEqual(
        str(ri),
        "ReadInstruction('train')",
    )
    self.check_from_spec('test', [('test', None, None)])
    # Addition of splits:
    self.check_from_spec(
        'train+test',
        [
            ('train', None, None),
            ('test', None, None),
        ],
    )
    # Absolute slicing:
    self.check_from_spec('train[0:0]', [('train', None, 0)])
    self.check_from_spec('train[:10]', [('train', None, 10)])
    self.check_from_spec('train[0:10]', [('train', None, 10)])
    self.check_from_spec('train[-10:]', [('train', 190, None)])
    self.check_from_spec('train[-100:-50]', [('train', 100, 150)])
    self.check_from_spec('train[-10:200]', [('train', 190, None)])
    self.check_from_spec('train[10:-10]', [('train', 10, 190)])
    self.check_from_spec('train[42:99]', [('train', 42, 99)])
    # Percent slicing, closest rounding:
    self.check_from_spec('train[:10%]', [('train', None, 20)])
    self.check_from_spec('train[90%:]', [('train', 180, None)])
    self.check_from_spec('train[-1%:]', [('train', 198, None)])
    ri = self.check_from_spec('test[:99%]', [('test', None, 100)])
    self.assertEqual(
        str(ri), "ReadInstruction('test[:99%]', rounding='closest')"
    )
    self.check_from_spec('train[:0.5%]', [('train', None, 1)])
    # No overlap:
    self.check_from_spec('test[100%:]', [('test', 101, None)])
    # Percent slicing, pct1_dropremainder rounding:
    ri = splits.ReadInstruction(
        'train', to=20, unit='%', rounding='pct1_dropremainder'
    )
    self.check_from_ri(ri, [('train', None, 40)])
    # test split has 101 examples.
    ri = splits.ReadInstruction(
        'test', to=100, unit='%', rounding='pct1_dropremainder'
    )
    self.check_from_ri(ri, [('test', None, 100)])
    # No overlap using 'pct1_dropremainder' rounding:
    ri1 = splits.ReadInstruction(
        'test', to=99, unit='%', rounding='pct1_dropremainder'
    )
    ri2 = splits.ReadInstruction(
        'test', from_=100, unit='%', rounding='pct1_dropremainder'
    )
    self.check_from_ri(ri1, [('test', None, 99)])
    self.check_from_ri(ri2, [('test', 100, None)])
    # Empty:
    # Slices resulting in empty datasets are valid with 'closest' rounding:
    self.check_from_spec('validation[:1%]', [('validation', None, 0)])
    # New integer syntax
    self.check_from_spec('train[4_2:9_9]', [('train', 42, 99)])
    self.check_from_spec('train[:1_0%]', [('train', None, 20)])

    # Supports splits with '-' in name.
    ri = self.check_from_spec('dev-train', [('dev-train', None, None)])

  def test_add(self):
    ri1 = splits.ReadInstruction.from_spec('train[10:20]')
    ri2 = splits.ReadInstruction.from_spec('test[10:20]')
    ri3 = splits.ReadInstruction.from_spec('train[1:5]')
    ri = ri1 + ri2 + ri3
    self.assertEqual(
        str(ri),
        (
            "ReadInstruction('train[10:20]')"
            "+ReadInstruction('test[10:20]')"
            "+ReadInstruction('train[1:5]')"
        ),
    )

  def test_invalid_rounding(self):
    with self.assertRaisesWithPredicateMatch(ValueError, 'Rounding should be'):
      splits.ReadInstruction('test', unit='%', rounding='unexisting')

  def test_invalid_unit(self):
    with self.assertRaisesWithPredicateMatch(ValueError, 'Unit should be'):
      splits.ReadInstruction('test', unit='kg', rounding='closest')

  def test_invalid_spec(self):
    # Invalid format:
    self.assertRaises(
        'validation[:250%:2]',
        "Unrecognized split format: 'validation[:250%:2]'",
    )
    # Unexisting split:
    self.assertRaises('imaginary', "Unknown split 'imaginary'")
    # Invalid boundaries abs:
    self.assertRaises('validation[:31]', 'incompatible with 30 examples')
    # Invalid boundaries %:
    self.assertRaises(
        'validation[:250%]', 'percent slice boundaries should be in [-100, 100]'
    )
    self.assertRaises(
        'validation[-101%:]',
        'percent slice boundaries should be in [-100, 100]',
    )
    # pct1_dropremainder with < 100 examples
    with self.assertRaisesWithPredicateMatch(
        ValueError, 'with less than 100 elements is forbidden'
    ):
      ri = splits.ReadInstruction(
          'validation', to=99, unit='%', rounding='pct1_dropremainder'
      )
      ri.to_absolute(self.splits)


class GetDatasetFilesTest(testing.TestCase):
  PATH_PATTERN = '/path/mnist-train.tfrecord-0000%d-of-00005'

  def _get_files(self, instruction):
    split_infos = [
        splits.SplitInfo(
            name='train',
            shard_lengths=[3, 2, 3, 2, 3],  # 13 examples.
            num_bytes=0,
            filename_template=_filename_template(
                dataset_name='mnist', split='train'
            ),
        )
    ]
    splits_dict = splits.SplitDict(split_infos=split_infos)
    return splits_dict[instruction].file_instructions

  def test_no_skip_no_take(self):
    instruction = splits._AbsoluteInstruction('train', None, None)
    files = self._get_files(instruction)
    expected_files = []
    for i, n in enumerate([3, 2, 3, 2, 3]):
      expected_files.append(
          shard_utils.FileInstruction(
              filename=self.PATH_PATTERN % i,
              skip=0,
              take=-1,
              examples_in_shard=n,
          )
      )
    self.assertEqual(files, expected_files)

  def test_skip(self):
    # One file is not taken, one file is partially taken.
    instruction = splits._AbsoluteInstruction('train', 4, None)
    files = self._get_files(instruction)
    self.assertEqual(
        files,
        [
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 1,
                skip=1,
                take=-1,
                examples_in_shard=2,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 2,
                skip=0,
                take=-1,
                examples_in_shard=3,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 3,
                skip=0,
                take=-1,
                examples_in_shard=2,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 4,
                skip=0,
                take=-1,
                examples_in_shard=3,
            ),
        ],
    )

  def test_take(self):
    # Two files are not taken, one file is partially taken.
    instruction = splits._AbsoluteInstruction('train', None, 6)
    files = self._get_files(instruction)
    self.assertEqual(
        files,
        [
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 0,
                skip=0,
                take=-1,
                examples_in_shard=3,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 1,
                skip=0,
                take=-1,
                examples_in_shard=2,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 2,
                skip=0,
                take=1,
                examples_in_shard=3,
            ),
        ],
    )

  def test_skip_take1(self):
    # A single shard with both skip and take.
    instruction = splits._AbsoluteInstruction('train', 1, 2)
    files = self._get_files(instruction)
    self.assertEqual(
        files,
        [
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 0,
                skip=1,
                take=1,
                examples_in_shard=3,
            ),
        ],
    )

  def test_skip_take2(self):
    # 2 elements in across two shards are taken in middle.
    instruction = splits._AbsoluteInstruction('train', 7, 9)
    files = self._get_files(instruction)
    self.assertEqual(
        files,
        [
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 2,
                skip=2,
                take=-1,
                examples_in_shard=3,
            ),
            shard_utils.FileInstruction(
                filename=self.PATH_PATTERN % 3,
                skip=0,
                take=1,
                examples_in_shard=2,
            ),
        ],
    )

  def test_touching_boundaries(self):
    # Nothing to read.
    instruction = splits._AbsoluteInstruction('train', 0, 0)
    files = self._get_files(instruction)
    self.assertEqual(files, [])

    instruction = splits._AbsoluteInstruction('train', None, 0)
    files = self._get_files(instruction)
    self.assertEqual(files, [])

    instruction = splits._AbsoluteInstruction('train', 3, 3)
    files = self._get_files(instruction)
    self.assertEqual(files, [])

    instruction = splits._AbsoluteInstruction('train', 13, None)
    files = self._get_files(instruction)
    self.assertEqual(files, [])

  def test_missing_shard_lengths(self):
    with self.assertRaisesWithPredicateMatch(ValueError, 'Shard empty.'):
      filename_template = _filename_template(
          split='train', dataset_name='mnist'
      )
      split_infos = [
          splits.SplitInfo(
              name='train',
              shard_lengths=[],
              num_bytes=0,
              filename_template=filename_template,
          ),
      ]
      splits_dict = splits.SplitDict(split_infos=split_infos)
      _ = splits_dict['train'].file_instructions


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""UnifiedQA datasets."""

from __future__ import annotations

import csv
import textwrap

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The UnifiedQA benchmark consists of 20 main question answering (QA) datasets
(each may have multiple versions) that target different formats as well as
various complex linguistic phenomena. These datasets are grouped into several
formats/categories, including: extractive QA, abstractive QA, multiple-choice
QA, and yes/no QA. Additionally, contrast sets are used for several datasets
(denoted with "contrast_sets_"). These evaluation sets are expert-generated
perturbations that deviate from the patterns common in the original dataset. For
several datasets that do not come with evidence paragraphs, two variants are
included: one where the datasets are used as-is and another that uses paragraphs
fetched via an information retrieval system as additional evidence, indicated
with "_ir" tags.

More information can be found at: https://github.com/allenai/unifiedqa.
"""

_HOMEPAGE = 'https://github.com/allenai/unifiedqa'

_CITATION = """@inproceedings{khashabi-etal-2020-unifiedqa,
    title = "{UNIFIEDQA}: Crossing Format Boundaries with a Single {QA} System",
    author = "Khashabi, Daniel  and
      Min, Sewon  and
      Khot, Tushar  and
      Sabharwal, Ashish  and
      Tafjord, Oyvind  and
      Clark, Peter  and
      Hajishirzi, Hannaneh",
    booktitle = "Findings of the Association for Computational Linguistics: EMNLP 2020",
    month = nov,
    year = "2020",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2020.findings-emnlp.171",
    doi = "10.18653/v1/2020.findings-emnlp.171",
    pages = "1896--1907",
}

Note that each UnifiedQA dataset has its own citation. Please see the source to
see the correct citation for each contained dataset."
"""

_URL_BASE = 'https://storage.googleapis.com/unifiedqa/data'


class UnifiedQAConfig(tfds.core.BuilderConfig):
  """BuilderConfig for UnifiedQA."""

  def __init__(self, *, features, data_urls, citation, header=False, **kwargs):
    """BuilderConfig for UnifiedQA.

    Args:
      features: `tfds.features.FeaturesDict`, specific feature dict for the
        dataset.
      data_urls: `dict`, urls to download the files from
      citation: `string`, citation for the data set
      header: `bool`, whether the data files have a header
      **kwargs: keyword arguments forwarded to super.
    """
    super(UnifiedQAConfig, self).__init__(**kwargs)
    self.features = features
    self.data_urls = data_urls
    self.citation = citation
    self.header = header


class UnifiedQA(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for UnifiedQA datasets."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  DEFAULT_FEATURES = tfds.features.FeaturesDict({
      'input': np.str_,
      'output': np.str_,
  })
  BUILDER_CONFIGS = [
      UnifiedQAConfig(
          name='ai2_science_elementary',
          description=textwrap.dedent(
              """\
          The AI2 Science Questions dataset consists of questions used in
          student assessments in the United States across elementary and middle
          school grade levels. Each question is 4-way multiple choice format and
          may or may not include a diagram element. This set consists of
          questions used for elementary school grade levels.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/ai2_science_elementary/train.tsv',
              'validation': f'{_URL_BASE}/ai2_science_elementary/dev.tsv',
              'test': f'{_URL_BASE}/ai2_science_elementary/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          http://data.allenai.org/ai2-science-questions
          """
          ),
      ),
      UnifiedQAConfig(
          name='ai2_science_middle',
          description=textwrap.dedent(
              """\
          The AI2 Science Questions dataset consists of questions used in
          student assessments in the United States across elementary and middle
          school grade levels. Each question is 4-way multiple choice format and
          may or may not include a diagram element. This set consists of
          questions used for middle school grade levels.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/ai2_science_middle/train.tsv',
              'validation': f'{_URL_BASE}/ai2_science_middle/dev.tsv',
              'test': f'{_URL_BASE}/ai2_science_middle/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          http://data.allenai.org/ai2-science-questions
          """
          ),
      ),
      UnifiedQAConfig(
          name='ambigqa',
          description=textwrap.dedent(
              """\
          AmbigQA is an open-domain question answering task which involves
          finding every plausible answer, and then rewriting the question for
          each one to resolve the ambiguity.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/ambigqa/train.tsv',
              'validation': f'{_URL_BASE}/ambigqa/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{min-etal-2020-ambigqa,
              title = "{A}mbig{QA}: Answering Ambiguous Open-domain Questions",
              author = "Min, Sewon  and
                Michael, Julian  and
                Hajishirzi, Hannaneh  and
                Zettlemoyer, Luke",
              booktitle = "Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)",
              month = nov,
              year = "2020",
              address = "Online",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/2020.emnlp-main.466",
              doi = "10.18653/v1/2020.emnlp-main.466",
              pages = "5783--5797",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_easy',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "easy" questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_easy/train.tsv',
              'validation': f'{_URL_BASE}/arc_easy/dev.tsv',
              'test': f'{_URL_BASE}/arc_easy/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_easy_dev',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "easy" questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_easy_dev/train.tsv',
              'validation': f'{_URL_BASE}/arc_easy_dev/dev.tsv',
              'test': f'{_URL_BASE}/arc_easy_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_easy_with_ir',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "easy" questions. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_easy_with_ir/train.tsv',
              'validation': f'{_URL_BASE}/arc_easy_with_ir/dev.tsv',
              'test': f'{_URL_BASE}/arc_easy_with_ir/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_easy_with_ir_dev',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "easy" questions. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_easy_with_ir_dev/train.tsv',
              'validation': f'{_URL_BASE}/arc_easy_with_ir_dev/dev.tsv',
              'test': f'{_URL_BASE}/arc_easy_with_ir_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_hard',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "hard" questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_hard/train.tsv',
              'validation': f'{_URL_BASE}/arc_hard/dev.tsv',
              'test': f'{_URL_BASE}/arc_hard/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_hard_dev',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "hard" questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_hard_dev/train.tsv',
              'validation': f'{_URL_BASE}/arc_hard_dev/dev.tsv',
              'test': f'{_URL_BASE}/arc_hard_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_hard_with_ir',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "hard" questions. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_hard_with_ir/train.tsv',
              'validation': f'{_URL_BASE}/arc_hard_with_ir/dev.tsv',
              'test': f'{_URL_BASE}/arc_hard_with_ir/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='arc_hard_with_ir_dev',
          description=textwrap.dedent(
              """\
          This dataset consists of genuine grade-school level, multiple-choice
          science questions, assembled to encourage research in advanced
          question-answering. The dataset is partitioned into a Challenge Set
          and an Easy Set, where the former contains only questions answered
          incorrectly by both a retrieval-based algorithm and a word
          co-occurrence algorithm. This set consists of "hard" questions. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/arc_hard_with_ir_dev/train.tsv',
              'validation': f'{_URL_BASE}/arc_hard_with_ir_dev/dev.tsv',
              'test': f'{_URL_BASE}/arc_hard_with_ir_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{clark2018think,
              title={Think you have solved question answering? try arc, the ai2 reasoning challenge},
              author={Clark, Peter and Cowhey, Isaac and Etzioni, Oren and Khot, Tushar and Sabharwal, Ashish and Schoenick, Carissa and Tafjord, Oyvind},
              journal={arXiv preprint arXiv:1803.05457},
              year={2018}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='boolq',
          description=textwrap.dedent(
              """\
          BoolQ is a question answering dataset for yes/no questions. These
          questions are naturally occurring ---they are generated in unprompted
          and unconstrained settings. Each example is a triplet of (question,
          passage, answer), with the title of the page as optional additional
          context. The text-pair classification setup is similar to existing
          natural language inference tasks.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/boolq/train.tsv',
              'validation': f'{_URL_BASE}/boolq/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{clark-etal-2019-boolq,
              title = "{B}ool{Q}: Exploring the Surprising Difficulty of Natural Yes/No Questions",
              author = "Clark, Christopher  and
                Lee, Kenton  and
                Chang, Ming-Wei  and
                Kwiatkowski, Tom  and
                Collins, Michael  and
                Toutanova, Kristina",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1300",
              doi = "10.18653/v1/N19-1300",
              pages = "2924--2936",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='boolq_np',
          description=textwrap.dedent(
              """\
          BoolQ is a question answering dataset for yes/no questions. These
          questions are naturally occurring ---they are generated in unprompted
          and unconstrained settings. Each example is a triplet of (question,
          passage, answer), with the title of the page as optional additional
          context. The text-pair classification setup is similar to existing
          natural language inference tasks. This version adds natural
          perturbations to the original version.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/boolq_np/train.tsv',
              'validation': f'{_URL_BASE}/boolq_np/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khashabi-etal-2020-bang,
              title = "More Bang for Your Buck: Natural Perturbation for Robust Question Answering",
              author = "Khashabi, Daniel  and
                Khot, Tushar  and
                Sabharwal, Ashish",
              booktitle = "Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP)",
              month = nov,
              year = "2020",
              address = "Online",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/2020.emnlp-main.12",
              doi = "10.18653/v1/2020.emnlp-main.12",
              pages = "163--170",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='commonsenseqa',
          description=textwrap.dedent(
              """\
          CommonsenseQA is a new multiple-choice question answering dataset that
          requires different types of commonsense knowledge to predict the
          correct answers . It contains questions with one correct answer and
          four distractor answers.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/commonsenseqa/train.tsv',
              'validation': f'{_URL_BASE}/commonsenseqa/dev.tsv',
              'test': f'{_URL_BASE}/commonsenseqa/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{talmor-etal-2019-commonsenseqa,
              title = "{C}ommonsense{QA}: A Question Answering Challenge Targeting Commonsense Knowledge",
              author = "Talmor, Alon  and
                Herzig, Jonathan  and
                Lourie, Nicholas  and
                Berant, Jonathan",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1421",
              doi = "10.18653/v1/N19-1421",
              pages = "4149--4158",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='commonsenseqa_test',
          description=textwrap.dedent(
              """\
          CommonsenseQA is a new multiple-choice question answering dataset that
          requires different types of commonsense knowledge to predict the
          correct answers . It contains questions with one correct answer and
          four distractor answers.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/commonsenseqa_test/train.tsv',
              'validation': f'{_URL_BASE}/commonsenseqa_test/dev.tsv',
              'test': f'{_URL_BASE}/commonsenseqa_test/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{talmor-etal-2019-commonsenseqa,
              title = "{C}ommonsense{QA}: A Question Answering Challenge Targeting Commonsense Knowledge",
              author = "Talmor, Alon  and
                Herzig, Jonathan  and
                Lourie, Nicholas  and
                Berant, Jonathan",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1421",
              doi = "10.18653/v1/N19-1421",
              pages = "4149--4158",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='contrast_sets_boolq',
          description=textwrap.dedent(
              """\
          BoolQ is a question answering dataset for yes/no questions. These
          questions are naturally occurring ---they are generated in unprompted
          and unconstrained settings. Each example is a triplet of (question,
          passage, answer), with the title of the page as optional additional
          context. The text-pair classification setup is similar to existing
          natural language inference tasks. This version uses contrast sets.
          These evaluation sets are expert-generated perturbations that deviate
          from the patterns common in the original dataset.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/contrast_sets_boolq/train.tsv',
              'validation': f'{_URL_BASE}/contrast_sets_boolq/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{clark-etal-2019-boolq,
              title = "{B}ool{Q}: Exploring the Surprising Difficulty of Natural Yes/No Questions",
              author = "Clark, Christopher  and
                Lee, Kenton  and
                Chang, Ming-Wei  and
                Kwiatkowski, Tom  and
                Collins, Michael  and
                Toutanova, Kristina",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1300",
              doi = "10.18653/v1/N19-1300",
              pages = "2924--2936",
          }
          """
          ),
          header=True,
      ),
      UnifiedQAConfig(
          name='contrast_sets_drop',
          description=textwrap.dedent(
              """\
          DROP is a crowdsourced, adversarially-created QA benchmark, in which a
          system must resolve references in a question, perhaps to multiple
          input positions, and perform discrete operations over them (such as
          addition, counting, or sorting). These operations require a much more
          comprehensive understanding of the content of paragraphs than what was
          necessary for prior datasets. This version uses contrast sets. These
          evaluation sets are expert-generated perturbations that deviate from
          the patterns common in the original dataset.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/contrast_sets_drop/train.tsv',
              'validation': f'{_URL_BASE}/contrast_sets_drop/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{dua-etal-2019-drop,
              title = "{DROP}: A Reading Comprehension Benchmark Requiring Discrete Reasoning Over Paragraphs",
              author = "Dua, Dheeru  and
                Wang, Yizhong  and
                Dasigi, Pradeep  and
                Stanovsky, Gabriel  and
                Singh, Sameer  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1246",
              doi = "10.18653/v1/N19-1246",
              pages = "2368--2378",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='contrast_sets_quoref',
          description=textwrap.dedent(
              """\
          This dataset tests the coreferential reasoning capability of reading
          comprehension systems. In this span-selection benchmark containing
          questions over paragraphs from Wikipedia, a system must resolve hard
          coreferences before selecting the appropriate span(s) in the
          paragraphs for answering questions. This version uses contrast sets.
          These evaluation sets are expert-generated perturbations that deviate
          from the patterns common in the original dataset.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/contrast_sets_quoref/train.tsv',
              'validation': f'{_URL_BASE}/contrast_sets_quoref/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{dasigi-etal-2019-quoref,
              title = "{Q}uoref: A Reading Comprehension Dataset with Questions Requiring Coreferential Reasoning",
              author = "Dasigi, Pradeep  and
                Liu, Nelson F.  and
                Marasovi{\'c}, Ana  and
                Smith, Noah A.  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing and the 9th International Joint Conference on Natural Language Processing (EMNLP-IJCNLP)",
              month = nov,
              year = "2019",
              address = "Hong Kong, China",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D19-1606",
              doi = "10.18653/v1/D19-1606",
              pages = "5925--5932",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='contrast_sets_ropes',
          description=textwrap.dedent(
              """\
          This dataset tests a system's ability to apply knowledge from a
          passage of text to a new situation. A system is presented a background
          passage containing a causal or qualitative relation(s) (e.g., "animal
          pollinators increase efficiency of fertilization in flowers"), a novel
          situation that uses this background, and questions that require
          reasoning about effects of the relationships in the background passage
          in the context of the situation. This version uses contrast sets.
          These evaluation sets are expert-generated perturbations that deviate
          from the patterns common in the original dataset.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/contrast_sets_ropes/train.tsv',
              'validation': f'{_URL_BASE}/contrast_sets_ropes/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lin-etal-2019-reasoning,
              title = "Reasoning Over Paragraph Effects in Situations",
              author = "Lin, Kevin  and
                Tafjord, Oyvind  and
                Clark, Peter  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2nd Workshop on Machine Reading for Question Answering",
              month = nov,
              year = "2019",
              address = "Hong Kong, China",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D19-5808",
              doi = "10.18653/v1/D19-5808",
              pages = "58--62",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='drop',
          description=textwrap.dedent(
              """\
          DROP is a crowdsourced, adversarially-created QA benchmark, in which a
          system must resolve references in a question, perhaps to multiple
          input positions, and perform discrete operations over them (such as
          addition, counting, or sorting). These operations require a much more
          comprehensive understanding of the content of paragraphs than what was
          necessary for prior datasets.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/drop/train.tsv',
              'validation': f'{_URL_BASE}/drop/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{dua-etal-2019-drop,
              title = "{DROP}: A Reading Comprehension Benchmark Requiring Discrete Reasoning Over Paragraphs",
              author = "Dua, Dheeru  and
                Wang, Yizhong  and
                Dasigi, Pradeep  and
                Stanovsky, Gabriel  and
                Singh, Sameer  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1246",
              doi = "10.18653/v1/N19-1246",
              pages = "2368--2378",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='mctest',
          description=textwrap.dedent(
              """\
          MCTest requires machines to answer multiple-choice reading
          comprehension questions about fictional stories, directly tackling the
          high-level goal of open-domain machine comprehension. Reading
          comprehension can test advanced abilities such as causal reasoning and
          understanding the world, yet, by being multiple-choice, still provide
          a clear metric. By being fictional, the answer typically can be found
          only in the story itself. The stories and questions are also carefully
          limited to those a young child would understand, reducing the world
          knowledge that is required for the task.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/mctest/train.tsv',
              'validation': f'{_URL_BASE}/mctest/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{richardson-etal-2013-mctest,
              title = "{MCT}est: A Challenge Dataset for the Open-Domain Machine Comprehension of Text",
              author = "Richardson, Matthew  and
                Burges, Christopher J.C.  and
                Renshaw, Erin",
              booktitle = "Proceedings of the 2013 Conference on Empirical Methods in Natural Language Processing",
              month = oct,
              year = "2013",
              address = "Seattle, Washington, USA",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D13-1020",
              pages = "193--203",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='mctest_corrected_the_separator',
          description=textwrap.dedent(
              """\
          MCTest requires machines to answer multiple-choice reading
          comprehension questions about fictional stories, directly tackling the
          high-level goal of open-domain machine comprehension. Reading
          comprehension can test advanced abilities such as causal reasoning and
          understanding the world, yet, by being multiple-choice, still provide
          a clear metric. By being fictional, the answer typically can be found
          only in the story itself. The stories and questions are also carefully
          limited to those a young child would understand, reducing the world
          knowledge that is required for the task.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/mctest_corrected_the_separator/train.tsv',
              'validation': (
                  f'{_URL_BASE}/mctest_corrected_the_separator/dev.tsv'
              ),
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{richardson-etal-2013-mctest,
              title = "{MCT}est: A Challenge Dataset for the Open-Domain Machine Comprehension of Text",
              author = "Richardson, Matthew  and
                Burges, Christopher J.C.  and
                Renshaw, Erin",
              booktitle = "Proceedings of the 2013 Conference on Empirical Methods in Natural Language Processing",
              month = oct,
              year = "2013",
              address = "Seattle, Washington, USA",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D13-1020",
              pages = "193--203",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='multirc',
          description=textwrap.dedent(
              """\
          MultiRC is a reading comprehension challenge in which questions can
          only be answered by taking into account information from multiple
          sentences. Questions and answers for this challenge were solicited and
          verified through a 4-step crowdsourcing experiment. The dataset
          contains questions for paragraphs across 7 different domains (
          elementary school science, news, travel guides, fiction stories, etc)
          bringing in linguistic diversity to the texts and to the questions
          wordings.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/multirc/train.tsv',
              'validation': f'{_URL_BASE}/multirc/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khashabi-etal-2018-looking,
              title = "Looking Beyond the Surface: A Challenge Set for Reading Comprehension over Multiple Sentences",
              author = "Khashabi, Daniel  and
                Chaturvedi, Snigdha  and
                Roth, Michael  and
                Upadhyay, Shyam  and
                Roth, Dan",
              booktitle = "Proceedings of the 2018 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long Papers)",
              month = jun,
              year = "2018",
              address = "New Orleans, Louisiana",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N18-1023",
              doi = "10.18653/v1/N18-1023",
              pages = "252--262",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='narrativeqa',
          description=textwrap.dedent(
              """\
          NarrativeQA is an English-lanaguage dataset of stories and
          corresponding questions designed to test reading comprehension,
          especially on long documents.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/narrativeqa/train.tsv',
              'validation': f'{_URL_BASE}/narrativeqa/dev.tsv',
              'test': f'{_URL_BASE}/narrativeqa/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{kocisky-etal-2018-narrativeqa,
              title = "The {N}arrative{QA} Reading Comprehension Challenge",
              author = "Ko{\v{c}}isk{\'y}, Tom{\'a}{\v{s}}  and
                Schwarz, Jonathan  and
                Blunsom, Phil  and
                Dyer, Chris  and
                Hermann, Karl Moritz  and
                Melis, G{\'a}bor  and
                Grefenstette, Edward",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "6",
              year = "2018",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q18-1023",
              doi = "10.1162/tacl_a_00023",
              pages = "317--328",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='narrativeqa_dev',
          description=textwrap.dedent(
              """\
          NarrativeQA is an English-lanaguage dataset of stories and
          corresponding questions designed to test reading comprehension,
          especially on long documents.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/narrativeqa_dev/train.tsv',
              'validation': f'{_URL_BASE}/narrativeqa_dev/dev.tsv',
              'test': f'{_URL_BASE}/narrativeqa_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{kocisky-etal-2018-narrativeqa,
              title = "The {N}arrative{QA} Reading Comprehension Challenge",
              author = "Ko{\v{c}}isk{\'y}, Tom{\'a}{\v{s}}  and
                Schwarz, Jonathan  and
                Blunsom, Phil  and
                Dyer, Chris  and
                Hermann, Karl Moritz  and
                Melis, G{\'a}bor  and
                Grefenstette, Edward",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "6",
              year = "2018",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q18-1023",
              doi = "10.1162/tacl_a_00023",
              pages = "317--328",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='natural_questions',
          description=textwrap.dedent(
              """\
          The NQ corpus contains questions from real users, and it requires QA
          systems to read and comprehend an entire Wikipedia article that may or
          may not contain the answer to the question. The inclusion of real user
          questions, and the requirement that solutions should read an entire
          page to find the answer, cause NQ to be a more realistic and
          challenging task than prior QA datasets.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/natural_questions/train.tsv',
              'validation': f'{_URL_BASE}/natural_questions/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='natural_questions_direct_ans',
          description=textwrap.dedent(
              """\
          The NQ corpus contains questions from real users, and it requires QA
          systems to read and comprehend an entire Wikipedia article that may or
          may not contain the answer to the question. The inclusion of real user
          questions, and the requirement that solutions should read an entire
          page to find the answer, cause NQ to be a more realistic and
          challenging task than prior QA datasets. This version consists of
          direct-answer questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/natural_questions_direct_ans/train.tsv',
              'validation': f'{_URL_BASE}/natural_questions_direct_ans/dev.tsv',
              'test': f'{_URL_BASE}/natural_questions_direct_ans/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='natural_questions_direct_ans_test',
          description=textwrap.dedent(
              """\
          The NQ corpus contains questions from real users, and it requires QA
          systems to read and comprehend an entire Wikipedia article that may or
          may not contain the answer to the question. The inclusion of real user
          questions, and the requirement that solutions should read an entire
          page to find the answer, cause NQ to be a more realistic and
          challenging task than prior QA datasets. This version consists of
          direct-answer questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': (
                  f'{_URL_BASE}/natural_questions_direct_ans_test/train.tsv'
              ),
              'validation': (
                  f'{_URL_BASE}/natural_questions_direct_ans_test/dev.tsv'
              ),
              'test': f'{_URL_BASE}/natural_questions_direct_ans_test/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='natural_questions_with_dpr_para',
          description=textwrap.dedent(
              """\
          The NQ corpus contains questions from real users, and it requires QA
          systems to read and comprehend an entire Wikipedia article that may or
          may not contain the answer to the question. The inclusion of real user
          questions, and the requirement that solutions should read an entire
          page to find the answer, cause NQ to be a more realistic and
          challenging task than prior QA datasets. This version includes
          additional paragraphs (obtained using the DPR retrieval engine) to
          augment each question.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/natural_questions_with_dpr_para/train.tsv',
              'validation': (
                  f'{_URL_BASE}/natural_questions_with_dpr_para/dev.tsv'
              ),
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='natural_questions_with_dpr_para_test',
          description=textwrap.dedent(
              """\
          The NQ corpus contains questions from real users, and it requires QA
          systems to read and comprehend an entire Wikipedia article that may or
          may not contain the answer to the question. The inclusion of real user
          questions, and the requirement that solutions should read an entire
          page to find the answer, cause NQ to be a more realistic and
          challenging task than prior QA datasets. This version includes
          additional paragraphs (obtained using the DPR retrieval engine) to
          augment each question.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': (
                  f'{_URL_BASE}/natural_questions_with_dpr_para_test/train.tsv'
              ),
              'test': (
                  f'{_URL_BASE}/natural_questions_with_dpr_para_test/test.tsv'
              ),
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='newsqa',
          description=textwrap.dedent(
              """\
          NewsQA is a challenging machine comprehension dataset of
          human-generated question-answer pairs. Crowdworkers supply questions
          and answers based on a set of news articles from CNN, with answers
          consisting of spans of text from the corresponding articles.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/newsqa/train.tsv',
              'validation': f'{_URL_BASE}/newsqa/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{trischler-etal-2017-newsqa,
              title = "{N}ews{QA}: A Machine Comprehension Dataset",
              author = "Trischler, Adam  and
                Wang, Tong  and
                Yuan, Xingdi  and
                Harris, Justin  and
                Sordoni, Alessandro  and
                Bachman, Philip  and
                Suleman, Kaheer",
              booktitle = "Proceedings of the 2nd Workshop on Representation Learning for {NLP}",
              month = aug,
              year = "2017",
              address = "Vancouver, Canada",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/W17-2623",
              doi = "10.18653/v1/W17-2623",
              pages = "191--200",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='openbookqa',
          description=textwrap.dedent(
              """\
          OpenBookQA aims to promote research in advanced question-answering,
          probing a deeper understanding of both the topic (with salient facts
          summarized as an open book, also provided with the dataset) and the
          language it is expressed in. In particular, it contains questions that
          require multi-step reasoning, use of additional common and commonsense
          knowledge, and rich text comprehension. OpenBookQA is a new kind of
          question-answering dataset modeled after open book exams for assessing
          human understanding of a subject.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/openbookqa/train.tsv',
              'validation': f'{_URL_BASE}/openbookqa/dev.tsv',
              'test': f'{_URL_BASE}/openbookqa/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{mihaylov-etal-2018-suit,
              title = "Can a Suit of Armor Conduct Electricity? A New Dataset for Open Book Question Answering",
              author = "Mihaylov, Todor  and
                Clark, Peter  and
                Khot, Tushar  and
                Sabharwal, Ashish",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
              month = oct # "-" # nov,
              year = "2018",
              address = "Brussels, Belgium",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D18-1260",
              doi = "10.18653/v1/D18-1260",
              pages = "2381--2391",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='openbookqa_dev',
          description=textwrap.dedent(
              """\
          OpenBookQA aims to promote research in advanced question-answering,
          probing a deeper understanding of both the topic (with salient facts
          summarized as an open book, also provided with the dataset) and the
          language it is expressed in. In particular, it contains questions that
          require multi-step reasoning, use of additional common and commonsense
          knowledge, and rich text comprehension. OpenBookQA is a new kind of
          question-answering dataset modeled after open book exams for assessing
          human understanding of a subject.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/openbookqa_dev/train.tsv',
              'validation': f'{_URL_BASE}/openbookqa_dev/dev.tsv',
              'test': f'{_URL_BASE}/openbookqa_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{mihaylov-etal-2018-suit,
              title = "Can a Suit of Armor Conduct Electricity? A New Dataset for Open Book Question Answering",
              author = "Mihaylov, Todor  and
                Clark, Peter  and
                Khot, Tushar  and
                Sabharwal, Ashish",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
              month = oct # "-" # nov,
              year = "2018",
              address = "Brussels, Belgium",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D18-1260",
              doi = "10.18653/v1/D18-1260",
              pages = "2381--2391",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='openbookqa_with_ir',
          description=textwrap.dedent(
              """\
          OpenBookQA aims to promote research in advanced question-answering,
          probing a deeper understanding of both the topic (with salient facts
          summarized as an open book, also provided with the dataset) and the
          language it is expressed in. In particular, it contains questions that
          require multi-step reasoning, use of additional common and commonsense
          knowledge, and rich text comprehension. OpenBookQA is a new kind of
          question-answering dataset modeled after open book exams for assessing
          human understanding of a subject. This version includes paragraphs
          fetched via an information retrieval system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/openbookqa_with_ir/train.tsv',
              'validation': f'{_URL_BASE}/openbookqa_with_ir/dev.tsv',
              'test': f'{_URL_BASE}/openbookqa_with_ir/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{mihaylov-etal-2018-suit,
              title = "Can a Suit of Armor Conduct Electricity? A New Dataset for Open Book Question Answering",
              author = "Mihaylov, Todor  and
                Clark, Peter  and
                Khot, Tushar  and
                Sabharwal, Ashish",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
              month = oct # "-" # nov,
              year = "2018",
              address = "Brussels, Belgium",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D18-1260",
              doi = "10.18653/v1/D18-1260",
              pages = "2381--2391",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='openbookqa_with_ir_dev',
          description=textwrap.dedent(
              """\
          OpenBookQA aims to promote research in advanced question-answering,
          probing a deeper understanding of both the topic (with salient facts
          summarized as an open book, also provided with the dataset) and the
          language it is expressed in. In particular, it contains questions that
          require multi-step reasoning, use of additional common and commonsense
          knowledge, and rich text comprehension. OpenBookQA is a new kind of
          question-answering dataset modeled after open book exams for assessing
          human understanding of a subject. This version includes paragraphs
          fetched via an information retrieval system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/openbookqa_with_ir_dev/train.tsv',
              'validation': f'{_URL_BASE}/openbookqa_with_ir_dev/dev.tsv',
              'test': f'{_URL_BASE}/openbookqa_with_ir_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{mihaylov-etal-2018-suit,
              title = "Can a Suit of Armor Conduct Electricity? A New Dataset for Open Book Question Answering",
              author = "Mihaylov, Todor  and
                Clark, Peter  and
                Khot, Tushar  and
                Sabharwal, Ashish",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
              month = oct # "-" # nov,
              year = "2018",
              address = "Brussels, Belgium",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D18-1260",
              doi = "10.18653/v1/D18-1260",
              pages = "2381--2391",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='physical_iqa',
          description=textwrap.dedent(
              """\
          This is a dataset for benchmarking progress in physical commonsense
          understanding. The underlying task is multiple choice question
          answering: given a question q and two possible solutions s1, s2, a
          model or a human must choose the most appropriate solution, of which
          exactly one is correct. The dataset focuses on everyday situations
          with a preference for atypical solutions. The dataset is inspired by
          instructables.com, which provides users with instructions on how to
          build, craft, bake, or manipulate objects using everyday materials.
          Annotators are asked to provide semantic perturbations or alternative
          approaches which are otherwise syntactically and topically similar to
          ensure physical knowledge is targeted. The dataset is further cleaned
          of basic artifacts using the AFLite algorithm.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/physical_iqa/train.tsv',
              'validation': f'{_URL_BASE}/physical_iqa/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{bisk2020piqa,
              title={Piqa: Reasoning about physical commonsense in natural language},
              author={Bisk, Yonatan and Zellers, Rowan and Gao, Jianfeng and Choi, Yejin and others},
              booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
              volume={34},
              number={05},
              pages={7432--7439},
              year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='qasc',
          description=textwrap.dedent(
              """\
          QASC is a question-answering dataset with a focus on sentence
          composition. It consists of 8-way multiple-choice questions about
          grade school science, and comes with a corpus of 17M sentences.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/qasc/train.tsv',
              'validation': f'{_URL_BASE}/qasc/dev.tsv',
              'test': f'{_URL_BASE}/qasc/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khot2020qasc,
              title={Qasc: A dataset for question answering via sentence composition},
              author={Khot, Tushar and Clark, Peter and Guerquin, Michal and Jansen, Peter and Sabharwal, Ashish},
              booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
              volume={34},
              number={05},
              pages={8082--8090},
              year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='qasc_test',
          description=textwrap.dedent(
              """\
          QASC is a question-answering dataset with a focus on sentence
          composition. It consists of 8-way multiple-choice questions about
          grade school science, and comes with a corpus of 17M sentences.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/qasc_test/train.tsv',
              'validation': f'{_URL_BASE}/qasc_test/dev.tsv',
              'test': f'{_URL_BASE}/qasc_test/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khot2020qasc,
              title={Qasc: A dataset for question answering via sentence composition},
              author={Khot, Tushar and Clark, Peter and Guerquin, Michal and Jansen, Peter and Sabharwal, Ashish},
              booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
              volume={34},
              number={05},
              pages={8082--8090},
              year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='qasc_with_ir',
          description=textwrap.dedent(
              """\
          QASC is a question-answering dataset with a focus on sentence
          composition. It consists of 8-way multiple-choice questions about
          grade school science, and comes with a corpus of 17M sentences. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/qasc_with_ir/train.tsv',
              'validation': f'{_URL_BASE}/qasc_with_ir/dev.tsv',
              'test': f'{_URL_BASE}/qasc_with_ir/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khot2020qasc,
              title={Qasc: A dataset for question answering via sentence composition},
              author={Khot, Tushar and Clark, Peter and Guerquin, Michal and Jansen, Peter and Sabharwal, Ashish},
              booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
              volume={34},
              number={05},
              pages={8082--8090},
              year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='qasc_with_ir_test',
          description=textwrap.dedent(
              """\
          QASC is a question-answering dataset with a focus on sentence
          composition. It consists of 8-way multiple-choice questions about
          grade school science, and comes with a corpus of 17M sentences. This
          version includes paragraphs fetched via an information retrieval
          system as additional evidence.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/qasc_with_ir_test/train.tsv',
              'validation': f'{_URL_BASE}/qasc_with_ir_test/dev.tsv',
              'test': f'{_URL_BASE}/qasc_with_ir_test/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{khot2020qasc,
              title={Qasc: A dataset for question answering via sentence composition},
              author={Khot, Tushar and Clark, Peter and Guerquin, Michal and Jansen, Peter and Sabharwal, Ashish},
              booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
              volume={34},
              number={05},
              pages={8082--8090},
              year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='quoref',
          description=textwrap.dedent(
              """\
          This dataset tests the coreferential reasoning capability of reading
          comprehension systems. In this span-selection benchmark containing
          questions over paragraphs from Wikipedia, a system must resolve hard
          coreferences before selecting the appropriate span(s) in the
          paragraphs for answering questions.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/quoref/train.tsv',
              'validation': f'{_URL_BASE}/quoref/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{dasigi-etal-2019-quoref,
              title = "{Q}uoref: A Reading Comprehension Dataset with Questions Requiring Coreferential Reasoning",
              author = "Dasigi, Pradeep  and
                Liu, Nelson F.  and
                Marasovi{\'c}, Ana  and
                Smith, Noah A.  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing and the 9th International Joint Conference on Natural Language Processing (EMNLP-IJCNLP)",
              month = nov,
              year = "2019",
              address = "Hong Kong, China",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D19-1606",
              doi = "10.18653/v1/D19-1606",
              pages = "5925--5932",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='race_string',
          description=textwrap.dedent(
              """\
          Race is a large-scale reading comprehension dataset. The dataset is
          collected from English examinations in China, which are designed for
          middle school and high school students. The dataset can be served as
          the training and test sets for machine comprehension.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/race_string/train.tsv',
              'validation': f'{_URL_BASE}/race_string/dev.tsv',
              'test': f'{_URL_BASE}/race_string/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lai-etal-2017-race,
              title = "{RACE}: Large-scale {R}e{A}ding Comprehension Dataset From Examinations",
              author = "Lai, Guokun  and
                Xie, Qizhe  and
                Liu, Hanxiao  and
                Yang, Yiming  and
                Hovy, Eduard",
              booktitle = "Proceedings of the 2017 Conference on Empirical Methods in Natural Language Processing",
              month = sep,
              year = "2017",
              address = "Copenhagen, Denmark",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D17-1082",
              doi = "10.18653/v1/D17-1082",
              pages = "785--794",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='race_string_dev',
          description=textwrap.dedent(
              """\
          Race is a large-scale reading comprehension dataset. The dataset is
          collected from English examinations in China, which are designed for
          middle school and high school students. The dataset can be served as
          the training and test sets for machine comprehension.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/race_string_dev/train.tsv',
              'validation': f'{_URL_BASE}/race_string_dev/dev.tsv',
              'test': f'{_URL_BASE}/race_string_dev/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lai-etal-2017-race,
              title = "{RACE}: Large-scale {R}e{A}ding Comprehension Dataset From Examinations",
              author = "Lai, Guokun  and
                Xie, Qizhe  and
                Liu, Hanxiao  and
                Yang, Yiming  and
                Hovy, Eduard",
              booktitle = "Proceedings of the 2017 Conference on Empirical Methods in Natural Language Processing",
              month = sep,
              year = "2017",
              address = "Copenhagen, Denmark",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D17-1082",
              doi = "10.18653/v1/D17-1082",
              pages = "785--794",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='ropes',
          description=textwrap.dedent(
              """\
          This dataset tests a system's ability to apply knowledge from a
          passage of text to a new situation. A system is presented a background
          passage containing a causal or qualitative relation(s) (e.g., "animal
          pollinators increase efficiency of fertilization in flowers"), a novel
          situation that uses this background, and questions that require
          reasoning about effects of the relationships in the background passage
          in the context of the situation.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/ropes/train.tsv',
              'validation': f'{_URL_BASE}/ropes/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lin-etal-2019-reasoning,
              title = "Reasoning Over Paragraph Effects in Situations",
              author = "Lin, Kevin  and
                Tafjord, Oyvind  and
                Clark, Peter  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2nd Workshop on Machine Reading for Question Answering",
              month = nov,
              year = "2019",
              address = "Hong Kong, China",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D19-5808",
              doi = "10.18653/v1/D19-5808",
              pages = "58--62",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='social_iqa',
          description=textwrap.dedent(
              """\
          This is a large-scale benchmark for commonsense reasoning about social
          situations. Social IQa contains multiple choice questions for probing
          emotional and social intelligence in a variety of everyday situations.
          Through crowdsourcing, commonsense questions along with correct and
          incorrect answers about social interactions are collected, using a new
          framework that mitigates stylistic artifacts in incorrect answers by
          asking workers to provide the right answer to a different but related
          question.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/social_iqa/train.tsv',
              'validation': f'{_URL_BASE}/social_iqa/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{sap-etal-2019-social,
              title = "Social {IQ}a: Commonsense Reasoning about Social Interactions",
              author = "Sap, Maarten  and
                Rashkin, Hannah  and
                Chen, Derek  and
                Le Bras, Ronan  and
                Choi, Yejin",
              booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing and the 9th International Joint Conference on Natural Language Processing (EMNLP-IJCNLP)",
              month = nov,
              year = "2019",
              address = "Hong Kong, China",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D19-1454",
              doi = "10.18653/v1/D19-1454",
              pages = "4463--4473",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='squad1_1',
          description=textwrap.dedent(
              """\
          This is a reading comprehension dataset consisting of questions posed
          by crowdworkers on a set of Wikipedia articles, where the answer to
          each question is a segment of text from the corresponding reading
          passage.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/squad1_1/train.tsv',
              'validation': f'{_URL_BASE}/squad1_1/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{rajpurkar-etal-2016-squad,
              title = "{SQ}u{AD}: 100,000+ Questions for Machine Comprehension of Text",
              author = "Rajpurkar, Pranav  and
                Zhang, Jian  and
                Lopyrev, Konstantin  and
                Liang, Percy",
              booktitle = "Proceedings of the 2016 Conference on Empirical Methods in Natural Language Processing",
              month = nov,
              year = "2016",
              address = "Austin, Texas",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D16-1264",
              doi = "10.18653/v1/D16-1264",
              pages = "2383--2392",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='squad2',
          description=textwrap.dedent(
              """\
          This dataset combines the original Stanford Question Answering Dataset
          (SQuAD) dataset with unanswerable questions written adversarially by
          crowdworkers to look similar to answerable ones.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/squad2/train.tsv',
              'validation': f'{_URL_BASE}/squad2/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{rajpurkar-etal-2018-know,
              title = "Know What You Don{'}t Know: Unanswerable Questions for {SQ}u{AD}",
              author = "Rajpurkar, Pranav  and
                Jia, Robin  and
                Liang, Percy",
              booktitle = "Proceedings of the 56th Annual Meeting of the Association for Computational Linguistics (Volume 2: Short Papers)",
              month = jul,
              year = "2018",
              address = "Melbourne, Australia",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/P18-2124",
              doi = "10.18653/v1/P18-2124",
              pages = "784--789",
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='winogrande_l',
          description=textwrap.dedent(
              """\
          This dataset is inspired by the original Winograd Schema Challenge
          design, but adjusted to improve both the scale and the hardness of the
          dataset. The key steps of the dataset construction consist of (1) a
          carefully designed crowdsourcing procedure, followed by (2) systematic
          bias reduction using a novel AfLite algorithm that generalizes
          human-detectable word associations to machine-detectable embedding
          associations. Training sets with differnt sizes are provided. This set
          corresponds to size `l`.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/winogrande_l/train.tsv',
              'validation': f'{_URL_BASE}/winogrande_l/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{sakaguchi2020winogrande,
            title={Winogrande: An adversarial winograd schema challenge at scale},
            author={Sakaguchi, Keisuke and Le Bras, Ronan and Bhagavatula, Chandra and Choi, Yejin},
            booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
            volume={34},
            number={05},
            pages={8732--8740},
            year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='winogrande_m',
          description=textwrap.dedent(
              """\
          This dataset is inspired by the original Winograd Schema Challenge
          design, but adjusted to improve both the scale and the hardness of the
          dataset. The key steps of the dataset construction consist of (1) a
          carefully designed crowdsourcing procedure, followed by (2) systematic
          bias reduction using a novel AfLite algorithm that generalizes
          human-detectable word associations to machine-detectable embedding
          associations. Training sets with differnt sizes are provided. This set
          corresponds to size `m`.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/winogrande_m/train.tsv',
              'validation': f'{_URL_BASE}/winogrande_m/dev.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{sakaguchi2020winogrande,
            title={Winogrande: An adversarial winograd schema challenge at scale},
            author={Sakaguchi, Keisuke and Le Bras, Ronan and Bhagavatula, Chandra and Choi, Yejin},
            booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
            volume={34},
            number={05},
            pages={8732--8740},
            year={2020}
          }
          """
          ),
      ),
      UnifiedQAConfig(
          name='winogrande_s',
          description=textwrap.dedent(
              """\
          This dataset is inspired by the original Winograd Schema Challenge
          design, but adjusted to improve both the scale and the hardness of the
          dataset. The key steps of the dataset construction consist of (1) a
          carefully designed crowdsourcing procedure, followed by (2) systematic
          bias reduction using a novel AfLite algorithm that generalizes
          human-detectable word associations to machine-detectable embedding
          associations. Training sets with differnt sizes are provided. This set
          corresponds to size `s`.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': f'{_URL_BASE}/winogrande_s/train.tsv',
              'validation': f'{_URL_BASE}/winogrande_s/dev.tsv',
              'test': f'{_URL_BASE}/winogrande_s/test.tsv',
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{sakaguchi2020winogrande,
            title={Winogrande: An adversarial winograd schema challenge at scale},
            author={Sakaguchi, Keisuke and Le Bras, Ronan and Bhagavatula, Chandra and Choi, Yejin},
            booktitle={Proceedings of the AAAI Conference on Artificial Intelligence},
            volume={34},
            number={05},
            pages={8732--8740},
            year={2020}
          }
          """
          ),
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(self.builder_config.features),
        supervised_keys=None,
        homepage=_HOMEPAGE,
        citation=self.builder_config.citation + '\n' + _CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    data_dir = dl_manager.download_and_extract(self.builder_config.data_urls)
    split_generators = {}

    if 'train' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.TRAIN: self._generate_examples(path=data_dir['train']),
          }
      )

    if 'validation' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.VALIDATION: self._generate_examples(
                  path=data_dir['validation']
              ),
          }
      )

    if 'test' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.TEST: self._generate_examples(path=data_dir['test']),
          }
      )
    return split_generators

  def _generate_examples(self, path):
    """Yields examples."""
    with epath.Path(path).open() as f:
      data = csv.reader(f, delimiter='\t')
      # Skip the header row
      if self.builder_config.header:
        next(data)
      for id_, row in enumerate(data):
        yield id_, {'input': row[0].strip(), 'output': row[1].strip()}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util to add metadata into an existing dataset folder."""
import os
from typing import List, Union

from etils import epath
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import splits as split_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.folder_dataset import compute_split_utils


def _get_file_infos(
    filename_template: naming.ShardedFileTemplate,
) -> List[naming.FilenameInfo]:
  """Returns the file infos from the files matching the given template."""
  file_infos = []
  for f in filename_template.data_dir.iterdir():
    file_info = filename_template.parse_filename_info(f.name)
    if file_info is not None:
      file_infos.append(file_info)
  if not file_infos:
    raise ValueError(
        f'Could not find data files in {filename_template.data_dir}. '
        f'Make sure to follow the pattern: `{filename_template.template}`'
    )
  return file_infos


def _construct_filename_template(
    data_dir: epath.PathLike,
    filename_template: Union[None, str, naming.ShardedFileTemplate] = None,
) -> naming.ShardedFileTemplate:
  """Returns a basic filename template based on the given data dir and template."""
  data_dir = epath.Path(data_dir)
  if filename_template is None:
    filename_template = naming.ShardedFileTemplate(
        data_dir=data_dir, template=naming.DEFAULT_FILENAME_TEMPLATE
    )
  elif isinstance(filename_template, str):
    filename_template = naming.ShardedFileTemplate(
        data_dir=data_dir, template=filename_template
    )
  return filename_template.replace(data_dir=data_dir)


def _enrich_filename_template(
    filename_template: naming.ShardedFileTemplate,
    file_infos: List[naming.FilenameInfo],
) -> naming.ShardedFileTemplate:
  """Enriches the given template with data from the given file infos."""
  # Use set with tuple expansion syntax to ensure all names are consistent.
  (dataset_name,) = {f.dataset_name for f in file_infos}
  if (
      filename_template.dataset_name
      and dataset_name
      and filename_template.dataset_name != dataset_name
  ):
    raise ValueError(
        f'Detected dataset name {dataset_name}, but '
        f'{filename_template.dataset_name} was specified '
        'in the filename template!'
    )
  elif dataset_name:
    # dataset_name can be None if the files don't specify the name
    filename_template = filename_template.replace(dataset_name=dataset_name)

  (filetype_suffix,) = {f.filetype_suffix for f in file_infos}
  if (
      filename_template.filetype_suffix
      and filename_template.filetype_suffix != filetype_suffix
  ):
    raise ValueError(
        f'Detected filetype suffix {filetype_suffix}, but '
        f'{filename_template.filetype_suffix} was specified '
        'in the filename template!'
    )
  elif filetype_suffix:
    # filetype_suffix can be None if the files don't specify the name
    filename_template = filename_template.replace(
        filetype_suffix=filetype_suffix
    )

  return filename_template


def write_metadata(
    *,
    data_dir: epath.PathLike,
    features: features_lib.feature.FeatureConnectorArg,
    split_infos: Union[None, epath.PathLike, List[split_lib.SplitInfo]] = None,
    version: Union[None, str, utils.Version] = None,
    filename_template: Union[None, str, naming.ShardedFileTemplate] = None,
    check_data: bool = True,
    **ds_info_kwargs,
) -> None:
  """Add metadata required to load with TFDS.

  See documentation for usage:
  https://www.tensorflow.org/datasets/external_tfrecord

  Args:
    data_dir: Dataset path on which to save the metadata
    features: dict of `tfds.features.FeatureConnector` matching the proto specs.
    split_infos: Can be either:  * A path to the pre-computed split info values
      ( the `out_dir` kwarg of `tfds.folder_dataset.compute_split_info`) * A
      list of `tfds.core.SplitInfo` (returned value of
      `tfds.folder_dataset.compute_split_info`) * `None` to auto-compute the
      split info.
    version: Optional dataset version (auto-infer by default, or fallback to
      1.0.0)
    filename_template: the template for the filenames of the data. If None, then
      the default template `'{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}'` is
      used. A string or a ShardedFileTemplate can be given for custom templates.
    check_data: If True, perform additional check to validate the data in
      data_dir is valid
    **ds_info_kwargs: Additional metadata forwarded to `tfds.core.DatasetInfo` (
      description, homepage,...). Will appear in the doc.
  """
  filename_template = _construct_filename_template(
      data_dir=data_dir, filename_template=filename_template
  )
  file_infos = _get_file_infos(filename_template=filename_template)
  filename_template = _enrich_filename_template(filename_template, file_infos)

  if version is None:  # Automatically detect the version
    if utils.Version.is_valid(filename_template.data_dir.name):
      version = filename_template.data_dir.name
    else:
      version = '1.0.0'

  dataset_identity = dataset_info.DatasetIdentity(
      name=filename_template.dataset_name,
      version=utils.Version(version),
      data_dir=os.fspath(filename_template.data_dir),
      module_name='',
  )

  # Create the metadata
  features = features_lib.features_dict.to_feature(features)
  ds_info = dataset_info.DatasetInfo(
      builder=dataset_identity,
      features=features,
      **ds_info_kwargs,
  )
  file_format = file_adapters.file_format_from_suffix(
      filename_template.filetype_suffix
  )
  ds_info.set_file_format(file_format)

  # Add the split infos
  split_dict = _load_splits(
      split_infos=split_infos,
      file_infos=file_infos,
      filename_template=filename_template,
  )
  ds_info.set_splits(split_dict)

  # Save all metadata (dataset_info.json, features.json,...)
  ds_info.write_to_directory(data_dir)

  # Make sure that the data can be loaded (feature connector matches the actual
  # specs)
  if check_data:
    utils.print_notebook(
        'Metadata written. Testing by reading first example. '
        'Set check_data=False to skip.'
    )
    builder = read_only_builder.builder_from_directory(data_dir)
    split_name = next(iter(builder.info.splits))
    (_,) = builder.as_dataset(
        split=f'{split_name}[:1]'
    )  # Load the first example


def _load_splits(
    *,
    split_infos: Union[None, epath.PathLike, List[split_lib.SplitInfo]],
    file_infos: List[naming.FilenameInfo],
    filename_template: naming.ShardedFileTemplate,
) -> split_lib.SplitDict:
  """Load the SplitDict which can be passed to DatasetInfo."""
  split_names = sorted(set(f.split for f in file_infos))

  if split_infos is None:  # Auto-compute the split-infos
    split_infos = compute_split_utils.compute_split_info(
        filename_template=filename_template
    )
  # Load the List[SplitInfo]
  elif isinstance(split_infos, epath.PathLikeCls):
    split_infos = compute_split_utils.split_infos_from_path(
        split_names=split_names,
        filename_template=filename_template,
    )
  elif all(isinstance(s, split_lib.SplitInfo) for s in split_infos):
    given_splits = sorted([s.name for s in split_infos])
    if given_splits != split_names:
      raise ValueError(f'Splits {given_splits} do not match {split_names}')
  else:
    raise TypeError(f'Invalid split info: {split_infos}')

  split_protos = [s.to_proto() for s in split_infos]
  return split_lib.SplitDict.from_proto(
      repeated_split_infos=split_protos,
      filename_template=filename_template,
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils common to Atari datasets."""

from __future__ import annotations

import dataclasses
from typing import Any, Dict

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_ATARI_DESCRIPTION = """
We are releasing a large and diverse dataset of gameplay following the protocol
described by [Agarwal et al., 2020](https://arxiv.org/abs/1907.04543), which can
be used to evaluate several discrete offline RL algorithms. The dataset is
generated by running an online DQN agent and recording transitions from its
replay during training with sticky actions
[Machado et al., 2018](https://arxiv.org/abs/1709.06009). As stated in
[Agarwal et al., 2020](https://arxiv.org/abs/1907.04543), for each game we use
data from five runs with 50 million transitions each. We release datasets for 46
Atari games. For details on how the dataset was generated, please refer to the
paper. Please see [this note](https://github.com/google-research/batch_rl#important-notes-on-atari-rom-versions)
about the ROM versions used to generate the datasets.

Atari is a standard RL benchmark. We recommend you to try offline RL methods on
Atari if you are interested in comparing your approach to other state of the art
offline RL methods with discrete actions.

The reward of each step is clipped (obtained with [-1, 1] clipping) and the 
episode includes the sum of the clipped reward per episode.
"""

_CITATION = """
@misc{gulcehre2020rl,
    title={RL Unplugged: Benchmarks for Offline Reinforcement Learning},
    author={Caglar Gulcehre and Ziyu Wang and Alexander Novikov and Tom Le Paine
        and  Sergio Gómez Colmenarejo and Konrad Zolna and Rishabh Agarwal and
        Josh Merel and Daniel Mankowitz and Cosmin Paduraru and Gabriel
        Dulac-Arnold and Jerry Li and Mohammad Norouzi and Matt Hoffman and
        Ofir Nachum and George Tucker and Nicolas Heess and Nando deFreitas},
    year={2020},
    eprint={2006.13888},
    archivePrefix={arXiv},
    primaryClass={cs.LG}
}
"""


def description():
  return _ATARI_DESCRIPTION


def citation():
  return _CITATION


@dataclasses.dataclass
class BuilderConfig(tfds.core.BuilderConfig):
  """Configuration of the task.

  Attributes:
    game: name of the Atari game
    run: name of the game run
  """

  game: str = 'Asterix'
  run: int = 1


_GAMES = [
    'Alien',
    'Amidar',
    'Assault',
    'Asterix',
    'Atlantis',
    'BankHeist',
    'BattleZone',
    'BeamRider',
    'Boxing',
    'Breakout',
    'Carnival',
    'Centipede',
    'ChopperCommand',
    'CrazyClimber',
    'DemonAttack',
    'DoubleDunk',
    'Enduro',
    'FishingDerby',
    'Freeway',
    'Frostbite',
    'Gopher',
    'Gravitar',
    'Hero',
    'IceHockey',
    'Jamesbond',
    'Kangaroo',
    'Krull',
    'KungFuMaster',
    'MsPacman',
    'NameThisGame',
    'Phoenix',
    'Pong',
    'Pooyan',
    'Qbert',
    'Riverraid',
    'RoadRunner',
    'Robotank',
    'Seaquest',
    'SpaceInvaders',
    'StarGunner',
    'TimePilot',
    'UpNDown',
    'VideoPinball',
    'WizardOfWor',
    'YarsRevenge',
    'Zaxxon',
]

_SHORT_GAMES = [
    'Carnival',
    'Gravitar',
    'StarGunner',
]


# Note that rewards and episode_return are actually also clipped.
def _feature_description():
  return {
      'checkpoint_idx': tf.io.FixedLenFeature([], tf.int64),
      'episode_idx': tf.io.FixedLenFeature([], tf.int64),
      'episode_return': tf.io.FixedLenFeature([], tf.float32),
      'clipped_episode_return': tf.io.FixedLenFeature([], tf.float32),
      'observations': tf.io.FixedLenSequenceFeature(
          [], tf.string, allow_missing=True
      ),
      'actions': tf.io.FixedLenSequenceFeature(
          [], tf.int64, allow_missing=True
      ),
      'unclipped_rewards': tf.io.FixedLenSequenceFeature(
          [], tf.float32, allow_missing=True
      ),
      'clipped_rewards': tf.io.FixedLenSequenceFeature(
          [], tf.float32, allow_missing=True
      ),
      'discounts': tf.io.FixedLenSequenceFeature(
          [], tf.float32, allow_missing=True
      ),
  }


def num_shards(game: str, shards: int) -> int:
  if game in _SHORT_GAMES:
    return shards - 1
  else:
    return shards


def builder_configs():
  configs = []
  for game in _GAMES:
    for run in range(1, 6):
      # pytype: disable=wrong-keyword-args
      configs.append(
          BuilderConfig(name=f'{game}_run_{run}', game=game, run=run)
      )
      # pytype: enable=wrong-keyword-args
  return configs


def atari_example_to_rlds(tf_example: tf.train.Example) -> Dict[str, Any]:
  """Generates an RLDS episode from an Atari TF Example.

  Args:
    tf_example: example from an Atari dataset.

  Returns:
    RLDS episode.
  """

  data = tf.io.parse_single_example(tf_example, _feature_description())
  episode_length = tf.size(data['actions'])
  is_first = tf.concat([[True], [False] * tf.ones(episode_length - 1)], axis=0)
  is_last = tf.concat([[False] * tf.ones(episode_length - 1), [True]], axis=0)

  is_terminal = [False] * tf.ones_like(data['actions'])
  discounts = data['discounts']
  if discounts[-1] == 0.0:
    is_terminal = tf.concat(
        [[False] * tf.ones(episode_length - 1, tf.int64), [True]], axis=0
    )
    # If the episode ends in a terminal state, in the last step only the
    # observation has valid information (the terminal state).
    discounts = tf.concat([discounts[1:], [0.0]], axis=0)
  episode = {
      # Episode Metadata
      'episode_id': data['episode_idx'],
      'checkpoint_id': data['checkpoint_idx'],
      'episode_return': data['episode_return'],
      'steps': {
          'observation': data['observations'],
          'action': data['actions'],
          'reward': data['unclipped_rewards'],
          'discount': discounts,
          'is_first': is_first,
          'is_last': is_last,
          'is_terminal': is_terminal,
      },
  }
  return episode


def file_prefix(prefix, run, game):
  return f'{prefix}/{game}/run_{run}'


def features_dict():
  return tfds.features.FeaturesDict({
      'steps': tfds.features.Dataset({
          'observation': tfds.features.Image(
              shape=(
                  84,
                  84,
                  1,
              ),
              dtype=np.uint8,
              encoding_format='png',
          ),
          'action': np.int64,
          'reward': tfds.features.Scalar(
              dtype=np.float32,
              doc=tfds.features.Documentation(
                  desc='Clipped reward.', value_range='[-1, 1]'
              ),
          ),
          'is_terminal': np.bool_,
          'is_first': np.bool_,
          'is_last': np.bool_,
          'discount': np.float32,
      }),
      'checkpoint_id': np.int64,
      'episode_id': np.int64,
      'episode_return': tfds.features.Scalar(
          dtype=np.float32,
          doc=tfds.features.Documentation(desc='Sum of the clipped rewards.'),
      ),
  })


def episode_id(episode):
  return f'{episode["checkpoint_id"]}_{episode["episode_id"]}'



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for registry."""
import os
import pathlib
import tempfile
import textwrap
from unittest import mock

import pytest

from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.community import register_package
from tensorflow_datasets.core.community import register_path
from tensorflow_datasets.core.community import registry as registry_lib
from tensorflow_datasets.core.utils import gcs_utils


class Ds0(testing.DummyDataset):
  pass


class Ds1(testing.DummyDataset):
  pass


@pytest.fixture(scope='module')
def dummy_register():
  """Dummy register."""
  with tempfile.TemporaryDirectory() as tmp_path:
    tmp_path = pathlib.Path(tmp_path)

    # Prepare the datasets
    # Namespace 0
    Ds0(data_dir=tmp_path / 'kaggle').download_and_prepare()
    Ds1(data_dir=tmp_path / 'kaggle2').download_and_prepare()
    # Namespace 1
    Ds0(data_dir=tmp_path / 'mlds').download_and_prepare()
    # Namespace 2: (non-existing)
    print('XXXXXXX: CONTENT!!!')
    for d in tmp_path.iterdir():
      print(d)
    for d in (tmp_path / 'kaggle').iterdir():
      print(d)
    for d in (tmp_path / 'kaggle2').iterdir():
      print(d)
    for d in (tmp_path / 'mlds').iterdir():
      print(d)

    content = textwrap.dedent(
        f"""
        [Namespaces]
        kaggle=[
            '{os.fspath(tmp_path / 'kaggle')}',
            '{os.fspath(tmp_path / 'kaggle2')}',
        ]
        mlds='{os.fspath(tmp_path / 'mlds')}'
        other='/tmp/path/to/non-existing-path'
        """
    )

    dummy_path = tmp_path / 'dummy-community-datasets.toml'
    dummy_path.write_text(content)
    yield registry_lib.DatasetRegistry(
        namespace_config=registry_lib.NamespaceConfig(config_path=dummy_path)
    )


def test_register_builder(dummy_register):  # pylint: disable=redefined-outer-name
  builder = dummy_register.builder(naming.DatasetName('kaggle:ds0'))
  assert 'kaggle' in builder.data_path.parts

  # Same dataset name can be loaded from different namespace
  builder = dummy_register.builder(naming.DatasetName('mlds:ds0'))
  assert 'mlds' in builder.data_path.parts

  builder = dummy_register.builder(
      naming.DatasetName('mlds:ds0'),
      data_dir=None,  # data_dir can be passed only if None
      version='1.0.0',
  )
  assert 'mlds' in builder.data_path.parts

  with pytest.raises(ValueError, match='`data_dir` cannot be set for'):
    dummy_register.builder(
        naming.DatasetName('mlds:ds0'), data_dir='/path/to/data_dir'
    )

  with pytest.raises(
      registered.DatasetNotFoundError, match='Namespace .* not found.'
  ):
    dummy_register.builder(naming.DatasetName('non-existing-namespace:ds0'))

  with pytest.raises(registered.DatasetNotFoundError):
    dummy_register.builder(naming.DatasetName('other:ds0'))


def test_register_path_list_builders(dummy_register):  # pylint: disable=redefined-outer-name
  assert dummy_register.list_builders() == [
      'kaggle:ds0',
      'kaggle:ds1',
      'mlds:ds0',
  ]


def test_load_register_for_path_github():
  registers = registry_lib._load_register_for_paths(
      namespace='huggingface',
      paths=['github://huggingface/datasets/tree/master/datasets'],
  )
  assert len(registers) == 1
  assert isinstance(registers[0], register_package.PackageRegister)
  assert registers[0]._path == gcs_utils.GCS_COMMUNITY_INDEX_PATH


def test_load_register_for_path_gcs():
  registers = registry_lib._load_register_for_paths(
      namespace='my_namespace',
      paths=['gs://my-bucket/datasets', 'gs://my-bucket2/datasets'],
  )
  assert len(registers) == 1
  assert isinstance(registers[0], register_path.DataDirRegister)


def test_load_register_for_path_mixed():
  with pytest.raises(RuntimeError, match='Both a path containing .*'):
    registry_lib._load_register_for_paths(
        namespace='my_namespace',
        paths=[
            'github://huggingface/datasets/tree/master/datasets',
            'gs://my-bucket/datasets',
        ],
    )


def test_community_register():
  assert 'huggingface' in registry_lib.community_register.list_namespaces()


def test_dataset_registry_list_builders():
  register1 = mock.create_autospec(register_path.DataDirRegister)
  register1.list_builders.return_value = ['a', 'b']
  register2 = mock.create_autospec(register_package.PackageRegister)
  register2.list_builders.return_value = ['c']
  namespace_config = mock.create_autospec(registry_lib.NamespaceConfig)
  namespace_config.registers_per_namespace.return_value = {
      'ns1': [register1],
      'ns2': [register2],
  }
  registry = registry_lib.DatasetRegistry(namespace_config=namespace_config)
  assert set(registry.list_namespaces()) == {'ns1', 'ns2'}
  assert set(registry.list_builders()) == {'a', 'b', 'c'}


def test_list_dataset_references():
  ref1 = naming.DatasetReference(dataset_name='ds1', namespace='kaggle')
  ref2 = naming.DatasetReference(dataset_name='ds2', namespace='kaggle')
  ref3 = naming.DatasetReference(dataset_name='ds3', namespace='kaggle')
  register1 = mock.create_autospec(register_path.DataDirRegister)
  register1.list_dataset_references.return_value = [ref1, ref2]
  register2 = mock.create_autospec(register_package.PackageRegister)
  register2.list_dataset_references.return_value = [ref3]
  namespace_config = mock.create_autospec(registry_lib.NamespaceConfig)
  namespace_config.registers_per_namespace.return_value = {
      'ns1': [register1],
      'ns2': [register2],
  }
  registry = registry_lib.DatasetRegistry(namespace_config=namespace_config)
  assert sorted(registry.list_dataset_references()) == [ref1, ref2, ref3]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_rwrl dataset."""

from __future__ import annotations

import collections
import dataclasses
import functools
import os
from typing import Any, Dict, Generator, Optional, Sequence, Text, Tuple, Union

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import rlu_common

_DESCRIPTION = """
Examples in the dataset represent SAR transitions stored when running a
partially online trained agent as described in https://arxiv.org/abs/1904.12901.
We follow the RLDS dataset format, as specified in
https://github.com/google-research/rlds#dataset-format.


We release 40 datasets on 8 tasks in total -- with no combined challenge and
easy combined challenge on the cartpole, walker, quadruped, and humanoid tasks.
Each task contains 5 different sizes of datasets, 1%, 5%, 20%, 40%, and 100%.
Note that the smaller dataset is not guaranteed to be a subset of the larger
ones. For details on how the dataset was generated, please refer to the paper.
"""

_CITATION = """
@misc{gulcehre2020rl,
    title={RL Unplugged: Benchmarks for Offline Reinforcement Learning},
    author={Caglar Gulcehre and Ziyu Wang and Alexander Novikov and Tom Le Paine
        and  Sergio Gómez Colmenarejo and Konrad Zolna and Rishabh Agarwal and
        Josh Merel and Daniel Mankowitz and Cosmin Paduraru and Gabriel
        Dulac-Arnold and Jerry Li and Mohammad Norouzi and Matt Hoffman and
        Ofir Nachum and George Tucker and Nicolas Heess and Nando deFreitas},
    year={2020},
    eprint={2006.13888},
    archivePrefix={arXiv},
    primaryClass={cs.LG}
}
"""

_HOMEPAGE = (
    'https://github.com/deepmind/deepmind-research/tree/master/rl_unplugged'
)

# Env constants
DOMAIN_NAMES = (
    'cartpole',
    'quadruped',
    'walker',
    'humanoid',
)
DOMAIN_TO_TASK = {
    'cartpole': 'swingup',
    'quadruped': 'walk',
    'walker': 'walk',
    'humanoid': 'walk',
}
# https://github.com/google-research/realworldrl_suite#rwrl-combined-challenge-benchmarks
COMBINED_CHALLENGES = (None, 'easy')
DATASET_SIZES = (
    '1_percent',
    '5_percent',
    '20_percent',
    '40_percent',
    '100_percent',
)
SHARDS_MAPPING = {
    # ('combined_challenge','domain','dataset_size'): number of shards.
    (None, 'cartpole', '1_percent'): 3,
    (None, 'cartpole', '5_percent'): 3,
    (None, 'cartpole', '20_percent'): 6,
    (None, 'cartpole', '40_percent'): 6,
    (None, 'cartpole', '100_percent'): 7,
    (None, 'quadruped', '1_percent'): 3,
    (None, 'quadruped', '5_percent'): 3,
    (None, 'quadruped', '20_percent'): 4,
    (None, 'quadruped', '40_percent'): 7,
    (None, 'quadruped', '100_percent'): 10,
    (None, 'walker', '1_percent'): 3,
    (None, 'walker', '5_percent'): 6,
    (None, 'walker', '20_percent'): 14,
    (None, 'walker', '40_percent'): 48,
    (None, 'walker', '100_percent'): 94,
    (None, 'humanoid', '1_percent'): 9,
    (None, 'humanoid', '5_percent'): 123,
    (None, 'humanoid', '20_percent'): 125,
    (None, 'humanoid', '40_percent'): 344,
    (None, 'humanoid', '100_percent'): 391,
    ('easy', 'cartpole', '1_percent'): 2,
    ('easy', 'cartpole', '5_percent'): 3,
    ('easy', 'cartpole', '20_percent'): 8,
    ('easy', 'cartpole', '40_percent'): 8,
    ('easy', 'cartpole', '100_percent'): 14,
    ('easy', 'quadruped', '1_percent'): 3,
    ('easy', 'quadruped', '5_percent'): 3,
    ('easy', 'quadruped', '20_percent'): 9,
    ('easy', 'quadruped', '40_percent'): 10,
    ('easy', 'quadruped', '100_percent'): 15,
    ('easy', 'walker', '1_percent'): 11,
    ('easy', 'walker', '5_percent'): 17,
    ('easy', 'walker', '20_percent'): 78,
    ('easy', 'walker', '40_percent'): 137,
    ('easy', 'walker', '100_percent'): 371,
    ('easy', 'humanoid', '1_percent'): 18,
    ('easy', 'humanoid', '5_percent'): 98,
    ('easy', 'humanoid', '20_percent'): 244,
    ('easy', 'humanoid', '40_percent'): 410,
    ('easy', 'humanoid', '100_percent'): 769,
}
# Control suite tasks have 1000 timesteps per episode. One additional timestep
# accounts for the very first observation where no action has been taken yet.
DEFAULT_NUM_TIMESTEPS = 1001
_DELIMITER = ':'


@dataclasses.dataclass
class BuilderConfig(tfds.core.BuilderConfig):
  """Configuration of the task.

  Attributes:
    domain: The RealWorldRL environment domain.
    task: The RealWorldRL environment task.
    combined_challenge: Combines multiple challenges into the same environment.
      https://github.com/google-research/realworldrl_suite#rwrl-combined-challenge-benchmarks
    dataset_size: the size of the dataset.
  """

  domain: str = 'cartpole'
  task: str = 'swingup'
  combined_challenge: Optional[str] = None
  dataset_size: str = '1_percent'


def _builder_configs():
  """Creates a list of default builder configs."""
  configs = []
  for combined_challenge, domain, dataset_size in SHARDS_MAPPING:
    task = DOMAIN_TO_TASK[domain]
    # pytype: disable=wrong-keyword-args
    configs.append(
        BuilderConfig(
            name=(
                f'{domain}_{task}_'
                f'combined_challenge_{str(combined_challenge).lower()}_'
                f'{dataset_size}'
            ),
            domain=domain,
            task=task,
            combined_challenge=combined_challenge,
            dataset_size=dataset_size,
        )
    )
    # pytype: enable=wrong-keyword-args
  return configs


def _decombine_key(k: str, delimiter: str = _DELIMITER) -> Sequence[str]:
  return k.split(delimiter)


def tf_example_to_feature_description(
    example: Union[tf.Tensor, bytes], num_timesteps=DEFAULT_NUM_TIMESTEPS
):
  """Takes a string tensor encoding an tf example and returns its features."""
  if isinstance(example, tf.Tensor):
    if not tf.executing_eagerly():
      raise AssertionError(
          'tf_example_to_feature_description() only works under eager mode.'
      )
    example = example.numpy()  # pytype: disable=attribute-error
  example = tf.train.Example.FromString(example)

  ret = {}
  for k, v in example.features.feature.items():
    l = len(v.float_list.value)
    if l % num_timesteps:
      raise ValueError(
          'Unexpected feature length %d. It should be divisible '
          'by num_timesteps: %d' % (l, num_timesteps)
      )
    size = l // num_timesteps
    ret[k] = tf.io.FixedLenFeature([num_timesteps, size], tf.float32)
  return ret


def tree_deflatten_with_delimiter(
    flat_dict: Dict[str, Any], delimiter: str = _DELIMITER
) -> Dict[str, Any]:
  """De-flattens a dict to its originally nested structure.

  Does the opposite of {combine_nested_keys(k) :v
                        for k, v in tree.flatten_with_path(nested_dicts)}
  Example: {'a:b': 1} -> {'a': {'b': 1}}
  Args:
    flat_dict: the keys of which equals the `path` separated by `delimiter`.
    delimiter: the delimiter that separates the keys of the nested dict.

  Returns:
    An un-flattened dict.
  """
  root = collections.defaultdict(dict)
  for delimited_key, v in flat_dict.items():
    keys = _decombine_key(delimited_key, delimiter=delimiter)
    node = root
    for k in keys[:-1]:
      node = node[k]
    node[keys[-1]] = v
  return dict(root)


def tf_feature_to_tfds_feature(
    nested: Union[tf.io.FixedLenFeature, Dict[Text, Any]]
):
  """Converts potentially nested tf features into tfds features."""
  if isinstance(nested, tf.io.FixedLenFeature):
    # For some reason the dicts are transposed within the tfds features.
    # See transpose_dict_list() under
    # tensorflow_datasets/core/features/sequence_feature.py
    return tfds.features.Tensor(shape=nested.shape[1:], dtype=nested.dtype)  # pytype: disable=attribute-error
  elif isinstance(nested, dict):
    ret = type(nested)()
    for k, v in nested.items():
      ret[k] = tf_feature_to_tfds_feature(v)
    return ret
  else:
    raise ValueError(f'Unsupported type {type(nested)}')


class RluRwrl(rlu_common.RLUBuilder):
  """DatasetBuilder for rlu_rwrl dataset."""

  VERSION = tfds.core.Version('1.0.1')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.0.1': (
          'Fixes a bug in RLU RWRL dataset where there are duplicated '
          'episode ids in one of the humanoid datasets.'
      ),
  }
  _INPUT_FILE_PREFIX = 'gs://rl_unplugged/rwrl/'

  BUILDER_CONFIGS = _builder_configs()

  def get_features_dict(self):
    # Loads the features dynamically.
    file_paths = rlu_common.get_files(
        prefix=self.get_file_prefix(), num_shards=self.num_shards()
    )

    # Take one item to get the output types and shapes.
    example_item = None
    iterator = tfds.as_numpy(tf.data.TFRecordDataset(file_paths[:1]))
    for example_item in iterator:
      break
    if example_item is None:
      raise ValueError('Empty dataset')

    feature_description = tf_example_to_feature_description(example_item)
    feature_description = tree_deflatten_with_delimiter(feature_description)
    return tfds.features.FeaturesDict({
        'steps': tfds.features.Dataset({
            'observation': tf_feature_to_tfds_feature(
                feature_description['observation']
            ),
            'action': tf_feature_to_tfds_feature(feature_description['action']),
            'reward': tf_feature_to_tfds_feature(feature_description['reward']),
            'is_terminal': np.bool_,
            'is_first': np.bool_,
            'is_last': np.bool_,
            'discount': tf_feature_to_tfds_feature(
                feature_description['discount']
            ),
        }),
        'episode_return': np.float32,
    })

  def get_description(self):
    return _DESCRIPTION

  def get_citation(self):
    return _CITATION

  def get_file_prefix(self):
    domain = self.builder_config.domain
    task = self.builder_config.task
    combined_challenge = self.builder_config.combined_challenge
    dataset_size = self.builder_config.dataset_size
    return (
        f'{self._INPUT_FILE_PREFIX}/'
        f'combined_challenge_{str(combined_challenge).lower()}/'
        f'{domain}/{task}/{dataset_size}/episodes.tfrecord'
    )

  def num_shards(self):
    try:
      return self._SHARDS  # For testing.  # type: ignore
    except AttributeError:
      pass
    domain = self.builder_config.domain
    combined_challenge = self.builder_config.combined_challenge
    dataset_size = self.builder_config.dataset_size
    return SHARDS_MAPPING[(combined_challenge, domain, dataset_size)]

  def tf_example_to_step_ds(  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
      self, tf_example: tf.Tensor, feature_description
  ) -> Dict[str, Any]:
    data = tf.io.parse_single_example(tf_example, feature_description)
    data = tree_deflatten_with_delimiter(data)

    # Shift the action and the rewards to conform to a common format.
    # The last action and reward will be a placeholder value.
    action = tf.concat((data['action'][1:], data['action'][:1]), axis=0)
    reward = tf.concat((data['reward'][1:], data['reward'][:1]), axis=0)
    discount = tf.concat((data['discount'][1:], data['discount'][:1]), axis=0)
    is_first = tf.concat(
        [[True], [False] * tf.ones(DEFAULT_NUM_TIMESTEPS - 1)], axis=0
    )
    is_last = tf.concat(
        [[False] * tf.ones(DEFAULT_NUM_TIMESTEPS - 1), [True]], axis=0
    )
    is_terminal = tf.squeeze(discount == 0.0, axis=-1)
    episode_return = tf.reduce_sum(reward)
    episode = {
        # Episode Metadata
        'episode_return': episode_return,
        'steps': {
            'observation': data['observation'],
            'action': action,
            'reward': reward,
            'discount': discount,
            'is_first': is_first,
            'is_last': is_last,
            'is_terminal': is_terminal,
        },
    }
    return episode

  def _generate_examples(self, paths):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam
    file_paths = paths['file_paths']

    # Take one item to get the output types and shapes.
    example_item = None
    for example_item in tf.data.TFRecordDataset(file_paths[:1]).take(1):
      break
    if example_item is None:
      raise ValueError('Empty dataset')

    feature_description = tf_example_to_feature_description(example_item)

    def _generate_examples_one_file(
        path,
    ) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
      """Yields examples from one file."""
      counter = 0
      key_prefix = os.path.basename(path)
      # Dataset of tf.Examples containing full episodes.
      example_ds = tf.data.TFRecordDataset(filenames=str(path))
      # Dataset of episodes, each represented as a dataset of steps.
      episode_ds = example_ds.map(
          functools.partial(
              self.tf_example_to_step_ds,
              feature_description=feature_description,
          ),
          num_parallel_calls=tf.data.experimental.AUTOTUNE,
      )
      episode_ds = tfds.as_numpy(episode_ds)
      for e in episode_ds:
        episode_id = counter
        yield f'{key_prefix}/{episode_id}', e
        counter += 1

    return beam.Create(file_paths) | beam.FlatMap(_generate_examples_one_file)
